# -*- coding: utf-8 -*-
"""
Created 2017/09/20
Latest update 2019/08/21

@author: Markus Eisenbach, Ronny Stricker

Description:
Helper functions for downloading datasets from FTP.
"""

from __future__ import print_function
import ftplib
try:
    import cPickle as pkl
except ImportError:
    import pickle as pkl
import hashlib
try:
    from pathlib import Path
except ImportError:
    from pathlib2 import Path  # python 2 backport
from .ftp_info import FTP_HOST, FTP_PORT, ftp_version_prefix
from .ftp_info import INFO_FILE_TEMPLATE, CHUNK_TEMPLATE, NP_FILE_TEMPLATE
from .ftp_info import dataset_info, dataset_checksums


def _cond_print(condition, content):
    if condition:
        print(content)


def _calc_md5(filename, debug_outputs=False):
    """Helper function to compute the MD5 checksum
       of the content of a file."""
    try:
        with filename.open('rb') as file_to_check:
            # read contents of the file
            data = file_to_check.read()
            # pipe contents of the file through
            return hashlib.md5(data).hexdigest()
    except IOError:
        _cond_print(debug_outputs, '[ERROR] cannot open file {} for computing md5 checksum'.format(filename))
        raise


def _is_valid(filename, md5_checksum, debug_outputs=False):
    """Helper function to check whether the MD5 checksum
       of the content of a file matches the expected checksum."""
    if filename.exists():
        file_md5 = _calc_md5(filename, debug_outputs)
        if file_md5 == md5_checksum:
            return True
        _cond_print(debug_outputs, '[ERROR] invalid checksum for file {}: {}'.format(filename, file_md5))
    else:
        _cond_print(debug_outputs, '[INFO] file {} does not exist'.format(filename))
    return False


def _read_pkl(file_path):
    with file_path.open('rb') as pkl_file:
        try:
            return pkl.load(pkl_file, encoding='latin1')  # python 3
        except TypeError:
            return pkl.load(pkl_file)  # python 2


def _db_subdir(version):
    if isinstance(version, str):
        return version
    return 'v{}'.format(version)


def connect_ftp(connection):
    if connection['ftp'] is None:
        connection['ftp'] = ftplib.FTP()
        connection['ftp'].connect(FTP_HOST, port=FTP_PORT)
        connection['ftp'].login(user=connection['user'], passwd=connection['password'])


def download_file(remote_path,
                  local_path,
                  checksum,
                  connection,
                  debug_outputs):
    if not _is_valid(local_path, checksum, debug_outputs):
        connect_ftp(connection)
        _cond_print(debug_outputs, '[INFO] downloading {}'.format(remote_path))

        for trial in range(3):
            # try to download file
            with local_path.open('wb') as dst_file:
                connection['ftp'].retrbinary('RETR '
                                             + ftp_version_prefix[connection['version']]
                                             + remote_path.as_posix(),
                                             dst_file.write)
            # verify checksum
            if _is_valid(local_path, checksum, debug_outputs):
                return
        raise IOError('[ERROR] Download of file {} failed three times in a row! Giving up.'.format(remote_path))


def _download_images(connection,
                     output_dir,
                     version=1,
                     subsets=None,
                     debug_outputs=False):

    # create database root dir
    ftp_dir = Path('images')
    db_dir = Path(output_dir) / _db_subdir(version) / ftp_dir
    db_dir.mkdir(parents=True, exist_ok=True)

    # download images
    images_file = 'images.zip'
    checksum = dataset_checksums[version]['images']['images']

    download_file(ftp_dir / images_file,
                  db_dir / images_file,
                  checksum,
                  connection,
                  debug_outputs)

    for subset in subsets:

        # download numpy file of subset
        np_file = NP_FILE_TEMPLATE.format(subset=subset)
        checksum = dataset_checksums[version]['images'][subset]

        download_file(ftp_dir / np_file,
                      db_dir / np_file,
                      checksum,
                      connection,
                      debug_outputs)


def _download_segmentation(connection,
                           output_dir,
                           version=2,
                           subsets=None,
                           debug_outputs=False):
    # create filelist
    files = ['images']
    if subsets is not None:
        files += subsets

    # create database root dir
    ftp_dir = Path('segmentation')
    db_dir = (Path(output_dir) / _db_subdir(version)) / ftp_dir
    db_dir.mkdir(parents=True, exist_ok=True)

    # download files
    for f in files:
        filename = f+".zip"
        checksum = dataset_checksums[version]['segmentation'][f]

        download_file(ftp_dir / filename,
                      db_dir / filename,
                      checksum,
                      connection,
                      debug_outputs)


def _download_patches(connection,
                      output_dir,
                      version=1,
                      patchsize=64,
                      issue='NORMvsDISTRESS',
                      subsets=None,
                      debug_outputs=False):

    # create database root dir
    ftp_dir = Path('{}_{}'.format(issue, patchsize))
    db_dir = Path(output_dir) / _db_subdir(version) / ftp_dir
    db_dir.mkdir(parents=True, exist_ok=True)

    for subset in subsets:
        # create subdir
        db_subdir = db_dir/subset
        ftp_subdir = ftp_dir/subset
        db_subdir.mkdir(exist_ok=True)

        # download subset info file
        sub_info_file = INFO_FILE_TEMPLATE.format(patchsize=str(patchsize),
                                                  issue=issue,
                                                  subset=subset)
        checksum = dataset_checksums[version][patchsize][issue][subset]

        download_file(ftp_subdir/sub_info_file,
                      db_subdir/sub_info_file,
                      checksum,
                      connection,
                      debug_outputs)

        # read subset info
        sub_info = _read_pkl(db_subdir/sub_info_file)

        # check if alternative ftp patch source is required
        if 'patch_issue' in dataset_checksums[version][patchsize][issue]:
            patch_issue = dataset_checksums[version][patchsize][issue]['patch_issue']
            ftp_patch_subdir = Path('{}_{}'.format(patch_issue, patchsize)) / subset
        else:
            patch_issue = issue
            ftp_patch_subdir = ftp_subdir

        # download all subset chunks
        for chunk_id in range(sub_info['n_chunks']):
            # use different issue and subdir for x files if required
            for data, d_issue, d_ftp_dir, checksum in zip(['x', 'y'],
                                                          [patch_issue, issue],
                                                          [ftp_patch_subdir, ftp_subdir],
                                                          sub_info['checksums'][chunk_id]):
                source_file = CHUNK_TEMPLATE.format(patchsize=str(patchsize),
                                                    issue=d_issue,
                                                    subset=subset,
                                                    part_nr=str(chunk_id).zfill(sub_info['chunk_descriptor_digits']),
                                                    data=data)
                target_file = CHUNK_TEMPLATE.format(patchsize=str(patchsize),
                                                    issue=issue,
                                                    subset=subset,
                                                    part_nr=str(chunk_id).zfill(sub_info['chunk_descriptor_digits']),
                                                    data=data)
                download_file(d_ftp_dir/source_file,
                              db_subdir/target_file,
                              checksum,
                              connection,
                              debug_outputs)


def download_dataset(
        login,
        output_dir,
        version=1,
        patchsize=64,
        issue='NORMvsDISTRESS',
        subsets=None,
        debug_outputs=False):

    if version not in dataset_info.keys():
        raise ValueError('Version must be one of {}'.format(dataset_info.keys()))
    if patchsize not in dataset_info[version]['patchsizes']:
        raise ValueError('Patchsize must be one of {}'.format(dataset_info[version]['patchsizes']))
    if issue not in dataset_info[version]['issues']:
        raise ValueError('Issue must be one of {} for selected version'.format(dataset_info[version]['issues']))

    if subsets is None:
        subsets = dataset_info[version]['subsets']

    if not isinstance(subsets, list):
        subsets = [subsets]

    unknown_subsets = [sub for sub in subsets if sub not in dataset_info[version]['subsets']]
    if len(unknown_subsets) > 0:
        raise ValueError('Subsets {} are not supported by the selected version. Supported subsets are: {}'.format(
            unknown_subsets,
            dataset_info[version]['subsets']
        ))

    if isinstance(patchsize, int):
        if dataset_checksums[version][patchsize][issue]['train'] is None:
            raise ValueError('Version, patchsize and subset combination available upon request only! '
                             'Please contact ronny.stricker@tu-ilmenau.de')

    try:
        split = int(login[-1])
    except ValueError:
        raise ValueError('Invalid login! Please provide proper login data.')

    connection = {
        'ftp': None,
        'user': login[:split],
        'password': login[split:-1],
        'version': version
    }

    if patchsize == 'images':
        _download_images(connection,
                         output_dir,
                         version,
                         subsets,
                         debug_outputs)
    elif patchsize == 'segmentation':
        _download_segmentation(connection,
                               output_dir,
                               version,
                               subsets,
                               debug_outputs)
    else:
        _download_patches(connection,
                          output_dir,
                          version,
                          patchsize,
                          issue,
                          subsets,
                          debug_outputs)

    # close ftp connection
    if connection['ftp'] is not None:
        connection['ftp'].quit()



"""
Created on 2019/09/11

@author: Ronny Stricker
Example for downloading patches of GAPs v1 dataset and usage.
"""

from gaps_dataset import gaps

# change these parameters:
destination_directory = '/your/download/destination/directory'
login = 'yourlogin'

# download patches (91 Gb!!)
gaps.download(login=login,
              output_dir=destination_directory,
              version=1,
              patchsize=64,                 # possible values: [64]
              issue='NORMvsDISTRESS',       # possible values: ['NORMvsDISTRESS']
              subsets=None,                 # possible values: [None, 'train', 'valid', 'test']
              debug_outputs=True)

# load training dataset info file
train_info = gaps.get_dataset_info(version=1,
                                   patchsize=64,
                                   issue='NORMvsDISTRESS',
                                   subset='train',
                                   datadir=destination_directory)

print("Training dataset info: {}".format(train_info))

# load all chunks of training dataset
for chunk_id in range(train_info['n_chunks']):
    x, y = gaps.load_chunk(chunk_id=chunk_id,
                           version=1,
                           patchsize=64,
                           issue='NORMvsDISTRESS',
                           subset='train',
                           datadir=destination_directory)

    # binary_label:
    #    0 = intact road,
    #    1 = distress

# replace train with valid or test respectively to extract patches of the other subsets



"""
Created on 2019/09/11

@author: Ronny Stricker
Example for downloading patches of GAPs v2 dataset and usage.
"""

from gaps_dataset import gaps

# change these parameters:
destination_directory = '/your/download/destination/directory'
login = 'yourlogin'

# download patches of the dataset
gaps.download(login=login,
              output_dir=destination_directory,
              version=2,
              patchsize=160,                # possible values: [64, 96, 128, 160, 192, 224, 256]
              issue='NORMvsDISTRESS_50k',   # possible values: ['NORMvsDISTRESS', 'NORMvsDISTRESS_50k', 'ZEB_50k']
              subsets=None,                 # possible values: [None, 'train', 'valid', 'valid-test', 'test']
              debug_outputs=True)

# load training dataset info file
train_info = gaps.get_dataset_info(version=2,
                                   patchsize=160,
                                   issue='NORMvsDISTRESS_50k',
                                   subset='train',
                                   datadir=destination_directory)

print("Training dataset info: {}".format(train_info))

# load all chunks of training dataset
for chunk_id in range(train_info['n_chunks']):
    x, y = gaps.load_chunk(chunk_id=chunk_id,
                           version=2,
                           patchsize=160,
                           issue='NORMvsDISTRESS_50k',
                           subset='train',
                           datadir=destination_directory)

    # binary_label:
    #    0 = intact road,
    #    1 = distress
    # class_label:
    #    0 = intact road,
    #    1 = applied patch,
    #    2 = pothole,
    #    3 = inlaid patch,
    #    4 = open joint,
    #    5 = crack

# replace train with valid or test respectively to extract patches of the other subsets



"""
Created on 2019/09/27

@author: Ronny Stricker
Example for creating patches from the patch location files.
"""

import os
import numpy as np
import cv2
import zipfile
import matplotlib.pyplot as plt
from gaps_dataset import gaps

#########

destination_directory = '/your/download/destination/directory'
login = 'yourlogin'
ds = 'test'
patchsize = 64

#########

# download images and segmentation masks
gaps.download(login=login,
              output_dir=destination_directory,
              version=2,
              patchsize='images',
              debug_outputs=True)

# unzip images
basedir = os.path.join(destination_directory, 'v2', 'images')
img_dir = os.path.join(basedir, 'images')

zip_filename = os.path.join(basedir, 'images.zip')
zip_ref = zipfile.ZipFile(zip_filename, 'r')
zip_ref.extractall(img_dir)
zip_ref.close()

patch_file = os.path.join(basedir, 'patch_references_{}.npz'.format(ds))
ps = patchsize / 2

# use variable old_img_name to avoid loading the input images for every single patch
old_img_id = None
img = None
patches = np.load(patch_file)['data']
for patch_id in range(patches.shape[0]):
    img_id, row, col, class_id = patches[patch_id]
    # load input image if it is different from the image used for the last patch
    if img_id != old_img_id:
        center = cv2.imread(os.path.join(img_dir, str(img_id).zfill(5) + '.jpg'))
        # add border to the image in order to draw patches without the need to check for image boundaries
        img = np.zeros((center.shape[0] + patchsize, center.shape[1] + patchsize, center.shape[2]), dtype=center.dtype)
        img[ps:ps + center.shape[0], ps:ps + center.shape[1]] = center
    old_img_id = img_id

    # draw patch from input image
    img_patch = img[row:row + patchsize, col:col + patchsize]
    assert img_patch.shape == (patchsize, patchsize, 3)

    # display patch
    plt.figure(num=str(class_id), figsize=(2, 2))
    plt.imshow(img_patch)
    plt.show()



"""
Created on 2021/08/12

@author: Ronny Stricker
Example for downloading segmentation masks of GAPs 10m dataset.
"""

from gaps_dataset import gaps
import zipfile
import os

# change these parameters:
destination_directory = '/your/download/destination/directory'
login = 'yourlogin'

# download images and segmentation masks
gaps.download(login=login,
              output_dir=destination_directory,
              version='10m',
              patchsize='segmentation',
              issue='ASFaLT',
              debug_outputs=True)

# unzip images and masks
basedir = os.path.join(destination_directory, '10m', 'segmentation')

for f in ['images', 'val_p']:
    zip_filename = os.path.join(basedir, f + '.zip')
    zip_ref = zipfile.ZipFile(zip_filename, 'r')
    zip_ref.extractall(os.path.join(basedir, f))
    zip_ref.close()

# Mask images are stored as grey value images.
# Coding of the grey values:
#    0 = VOID
#    1 = Inlaid patch
#    2 = Applied patch
#    3 = Sealed crack
#    4 = Crack
#    5 = Open joint
#    6 = Pothole
#    7 = Raveling
#    8 = Scratch
#    9 = Bleeding
#   10 = Road marking
#   11 = Surface water drain
#   12 = Manhole
#   13 = Expansion Joint
#   14 = Curb
#   15 = Cobblestone
#   16 = Drill hole
#   17 = Object mobile
#   18 = Object fixed
#   19 = Joint
#   20 = Road verge
#   21 = Vegetation
#   22 = Induction loop
#   23 = Normal



# -*- coding: utf-8 -*-
"""
Created 2017/04/27
Last update 2019/09/11

@author: Markus Eisenbach, Ronny Stricker

Description:
Automatic download of GAPs dataset if not available on PC.
"""

from __future__ import print_function, absolute_import
import numpy as np
from .ftp_utils import download_dataset, _read_pkl
from .ftp_info import INFO_FILE_TEMPLATE, CHUNK_TEMPLATE
try:
    from pathlib import Path
except ImportError:
    from pathlib2 import Path  # python 2 backport


def download(
        login,
        output_dir,
        version=1,
        patchsize=64,
        issue='NORMvsDISTRESS',
        subsets=None,
        debug_outputs=False):
    """
    Downloads the GAPs dataset with the specified patch size and issue.
    Please provide a valid login.

    Parameters
    ----------
    login : str
        Please enter the login you have received by mail.
        Note that you required different logins for the different versions of the dataset.
    output_dir : str
        Specify the directory where the dataset should be stored.
        The download script will create different subdirs for the different versions and issues.
    version : int
        The requested version of the dataset. Possible values are [1, 2].
    patchsize : int
        The requested patch size.
    issue : str
        The requested issue of the dataset. The issues differ in terms of the number of classes.
    subsets : list of str
        Specify the desired subsets of the dataset.
        All available subsets will be downloaded if None is passed as argument.
    debug_outputs : bool
        Specifies if debug outputs are desired.

    """
    download_dataset(login, output_dir, version, patchsize, issue, subsets, debug_outputs)


def get_dataset_info(version=1,
                     patchsize=64,
                     issue='NORMvsDISTRESS',
                     subset='train',
                     datadir='/local/datasets/gaps'):
    """
    Get information dictionary for a single subset of the dataset.
    Please note that the specified dataset has to be downloaded first.

    Parameters
    ----------
    version : int
        The requested version of the dataset. Possible values are [1, 2].
    patchsize : int
        The requested patch size.
    issue : str
        The requested issue of the dataset. The issues differ in terms of the number of classes.
    subset : list of str
        Specify the desired subsets of the dataset.
        All available subsets will be downloaded if None is passed as argument.
    datadir : str
        Specify the directory where the dataset is located (provide the same dir used in the download function call).

    Returns
    -------
    dict
        Dictionary with dataset info.

    """

    # obtain dataset info
    dataset_path = Path(datadir) / 'v{}'.format(version) / '{}_{}'.format(issue, patchsize) / subset
    dataset_info_file = dataset_path / INFO_FILE_TEMPLATE.format(patchsize=str(patchsize),
                                                                 issue=issue,
                                                                 subset=subset)
    if not dataset_info_file.exists():
        msg = 'cannot find dataset info file!'
        msg2 = 'Please download the dataset first.'
        msg3 = 'See http://www.tu-ilmenau.de/neurob/data-sets-code/gaps/'
        raise IOError('[ERROR] {}\n       HINT: {}\n       HINT: {}'.format(msg, msg2, msg3))

    return _read_pkl(dataset_info_file)


def load_chunk(chunk_id,
               version=1,
               patchsize=64,
               issue='NORMvsDISTRESS',
               subset='train',
               datadir='/local/datasets/gaps'):
    """
    Get information dictionary for a single subset of the dataset.
    Please note that the specified dataset has to be downloaded first.

    Parameters
    ----------
    chunk_id : int
        The id of the desired chunk. E.g. 0, 1, 2, ...
    version : int
        The requested version of the dataset. Possible values are [1, 2].
    patchsize : int
        The requested patch size.
    issue : str
        The requested issue of the dataset. The issues differ in terms of the number of classes.
    subset : list of str
        Specify the desired subsets of the dataset.
        All available subsets will be downloaded if None is passed as argument.
    datadir : str
        Specify the directory where the dataset is located (provide the same dir used in the download function call).

    Returns
    -------
    (ndarray, ndarray)
        The first mat has four dimensions (sample, channel, y, x) and contains the patches.
        The second mat has one dimension and contains the target values for every single sample.

    """

    # obtain dataset info
    ds_info = get_dataset_info(version,
                               patchsize,
                               issue,
                               subset,
                               datadir)

    if chunk_id < 0 or chunk_id >= ds_info['n_chunks']:
        raise ValueError("Chunk with id {} not available. The total number of chunks is {}.".format(
            chunk_id,
            ds_info['n_chunks']))

    # continue extracting the chunk
    dataset_path = Path(datadir) / 'v{}'.format(version) / '{}_{}'.format(issue, patchsize) / subset
    x_file, y_file = (dataset_path / CHUNK_TEMPLATE.format(
        patchsize=str(patchsize),
        issue=issue,
        subset=subset,
        part_nr=str(chunk_id).zfill(ds_info['chunk_descriptor_digits']),
        data=data) for data in ['x', 'y'])
    try:
        return np.load(x_file), np.load(y_file)
    except IOError:
        print('[ERROR] Failed to load chunk {}'.format(chunk_id))
        raise



"""
Created on 2019/09/11

@author: Ronny Stricker
Example for downloading segmentation masks of GAPs v2 dataset.
"""

from gaps_dataset import gaps
import zipfile
import os

# change these parameters:
destination_directory = '/your/download/destination/directory'
login = 'yourlogin'

# download images and segmentation masks
gaps.download(login=login,
              output_dir=destination_directory,
              version=2,
              patchsize='segmentation',
              debug_outputs=True)

# unzip images and masks
basedir = os.path.join(destination_directory, 'v2', 'segmentation')

for f in ['images', 'train', 'valid', 'test', 'valid-test']:
    zip_filename = os.path.join(basedir, f + '.zip')
    zip_ref = zipfile.ZipFile(zip_filename, 'r')
    zip_ref.extractall(os.path.join(basedir, f))
    zip_ref.close()

# Mask images are stored as grey value images.
# Coding of the grey values:
#    0 = VOID
#    1 = intact road,
#    2 = applied patch,
#    3 = pothole,
#    4 = inlaid patch,
#    5 = open joint,
#    6 = crack
#    7 = street inventory
